# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution - module extension
#    Copyright (C) 2010- O4SB (<http://openforsmallbusiness.co.nz>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
'''Implements central billing for Openerp by changing the partner at invoice time'''

from osv import osv, fields
from tools.translate import _

class res_partner(osv.osv):
    '''inherit base.res_partner and add columns to allow central invoicing'''
    _inherit = 'res.partner'
    
    _columns = {
            'central_inv': fields.boolean('Central Billing', 
                                          help='When enabled invoices for subsidiaries'\
                                            'will be billed to the parent account'),
            'store_ref': fields.char('Store Code', size=16, 
                                     help='If the customer requires specific store '
                                        'references on documentation'),
        }
    
    def copy(self, cr, uid, id, defaults, context=None):
        '''due to duplicate constraint we must set store_ref to False when
        copying:
        @return: the copied object'''
        defaults.update({'store_ref': False})
        return super(res_partner, self).copy(cr, uid, id, defaults, context)
    
    def _check_store_code(self, cr, uid, ids, context=None):
        """This function checks that the store code is unique within 
        the account hierarchy it belongs"""
        partner_obj = self.pool.get('res.partner')
        partner = partner_obj.browse(cr, uid, ids[0], context)
        if partner.store_ref:
            if partner.parent_id:
                group_partners = partner_obj.search(cr, uid, 
                                                    [('id', 'child_of', [partner.parent_id.id])])
            else:
                group_partners = partner_obj.search(cr, uid, [('id', 'child_of', [partner.id])])
            group_partners = list(set(group_partners) - set([ids[0]]))
            if group_partners:
                for part in partner_obj.browse(cr, uid, group_partners, context):
                    if part.store_ref and part.store_ref == partner.store_ref:
                        return False
        return True

    
    _constraints = [
        (_check_store_code, _('Store Codes must be unique per group'), ['store_ref']),
    ]
    
res_partner()

class account_invoice(osv.osv):
    '''inherits account.account_invoice and adds the order_partner_id field
    as well as overriding ORM functions to ensure the parent partner and order partner
    are written and created correctly'''  
    _inherit = 'account.invoice'
    
    _columns = {
            'order_partner_id': fields.many2one('res.partner', 'Ordering Partner', 
                                                required=False, readonly=True),
            }
    
    def create(self, cr, uid, vals, context=None):
        '''Function overrides create to ensure that parent account is always used'''
        if vals.get('partner_id', False) and \
                        vals.get('type', False) not in ['in_invoice', 'in_refund']:
            partner = self.pool.get('res.partner').browse(cr, uid, vals['partner_id'], 
                                                          context=context)
            if partner.parent_id and partner.parent_id.central_inv:
                vals.update({'partner_id': partner.parent_id.id,
                             'order_partner_id': partner.id, })
        return super(account_invoice, self).create(cr, uid, vals, context)
    
    def write(self, cr, uid, ids, vals, context=None):
        '''Function overrides create to ensure that parent account is always used'''
        if vals.get('partner_id', False):
            partner = self.pool.get('res.partner').browse(cr, uid, vals['partner_id'], 
                                                          context=context)
            if partner.parent_id and partner.parent_id.central_inv:
                new_vals_to_write = vals.copy()
                new_vals_to_write.update({'partner_id': partner.parent_id.id,
                                            'order_partner_id': partner.id, })
                new_ids_to_write = []
                for inv in self.browse(cr, uid, ids, context):
                    if inv.type in ['out_invoice', 'out_refund']:
                        new_ids_to_write.append(inv.id)
                if new_ids_to_write:
                    self.write(cr, uid, new_ids_to_write, new_vals_to_write, context)
                    ids = list(set(ids) - set(new_ids_to_write))
        if ids:
            return super(account_invoice, self).write(cr, uid, ids, vals, context)
        return True
    
    def search(self, cr, uid, args, offset=0,
                limit=None, order=None,
                context=None, count=False):
        '''override search so we find subsidiary invoices when looking at that partner
        @note: this could break quite easily if used with multiple custom filters'''
        for search_args in args:
            if search_args[0] == 'partner_id' and search_args[1] in ['=', 'like', 'ilike']:
                args.remove(search_args)
                args.extend(['|', search_args, 
                             ('order_partner_id', search_args[1], search_args[2])] )
                break
        return super(account_invoice, self).search(cr, uid, args, offset, limit, order,
                                                   context, count)
        
    def refund(self, cr, uid, ids, date=None, period_id=None, description=None, journal_id=None):
        '''overide refund to make sure we preserve the original ordering partner, also we 
        update the credit not to display the invoice in the origin and the comments.  To work
        either the refunds must be for a single original order partner or all order partners. In
        practice this is not an issue as the wizard only works on a single id, but at least
        this way we are covered if that changes.'''
        order_partner_ids = set([x.order_partner_id and x.order_partner_id.id or False
                                    for x in self.browse(cr, uid, ids)])
        if len(order_partner_ids) != 1:
            raise osv.except_osv(_('Error:'), 
                                 _('May only refund centrally billed invoices for one store at a time'))
        new_ids = super(account_invoice, self).refund(cr, uid, ids, date, period_id,
                                                      description, journal_id)
        vals_to_write = {'order_partner_id': list(order_partner_ids)[0]}
        if len(ids) == 1:
            inv_num = self.pool.get('account.invoice').browse(cr, uid, ids)[0].number
            vals_to_write.update({'comment': 'Credit Raised against Invoice %s' % inv_num,
                                  'origin': inv_num})
        self.write(cr, uid, new_ids, vals_to_write)
        
            
        return new_ids

account_invoice()

class account_invoice_refund(osv.osv_memory):
    _inherit = 'account.invoice.refund'
    
    def invoice_refund(self, cr, uid, ids, context=None):
        '''All this horrible shit just so we can call super, inefficient
        but don't blame me, parent module needs refactoring!!! In essence
        we parse the return domain for active ids (which is broken anyway)
        and add comments and origins to the invoice.
        @return: returns a view object being the credit note created'''
        res = super(account_invoice_refund, self).invoice_refund(cr, uid, ids, context)
        inv_obj = self.pool.get('account.invoice')
        if self.read(cr, uid, ids, [], context)[0]['filter_refund'] == 'modify' and \
                    len(context.get('active_ids')) == 1:
            for args in res['domain']:
                if args[0] == 'id':
                    for inv in inv_obj.browse(cr, uid, args[2]):
                        if inv.type == 'out_invoice':
                            orig_inv = inv_obj.browse(cr, uid, context.get('active_ids'))[0]
                            inv.write({'comment': 'Original Invoice %s' % orig_inv.number,
                                       'order_partner_id': orig_inv.order_partner_id.id,
                                       'origin': orig_inv.number })
                            break
                    break
        return res

account_invoice_refund()
